from __future__ import annotations

from typing import Any

class Body:
    body_type: str
    content: Any

    def __init__(self, body_type: str, content: Any = None) -> None: ...

class HttpRequest:
    method: str
    uri: str
    headers: dict[str, list[str]]
    body: Body

    def __init__(
        self,
        method: str,
        uri: str,
        headers: dict[str, list[str]] | None = None,
        body: Body | None = None,
    ) -> None: ...

class HttpResponse:
    status: int
    headers: dict[str, list[str]]
    body: Body

    def __init__(
        self,
        status: int,
        headers: dict[str, list[str]] | None = None,
        body: Body | None = None,
    ) -> None: ...

class HttpInteraction:
    request: HttpRequest
    response: HttpResponse
    recorded_at: str

    def __init__(self, request: HttpRequest, response: HttpResponse, recorded_at: str) -> None: ...

class GrpcRequest:
    method: str
    metadata: dict[str, list[str]]
    body: Body

    def __init__(
        self,
        method: str,
        metadata: dict[str, list[str]] | None = None,
        body: Body | None = None,
    ) -> None: ...

class GrpcResponse:
    status_code: int
    status_message: str
    metadata: dict[str, list[str]]
    body: Body

    def __init__(
        self,
        status_code: int,
        status_message: str | None = None,
        metadata: dict[str, list[str]] | None = None,
        body: Body | None = None,
    ) -> None: ...

class GrpcInteraction:
    request: GrpcRequest
    response: GrpcResponse
    json_debug: Any
    recorded_at: str

    def __init__(
        self,
        request: GrpcRequest,
        response: GrpcResponse,
        recorded_at: str,
        json_debug: Any = None,
    ) -> None: ...

class WsFrame:
    direction: str
    frame_type: str
    body: Body
    offset_ms: int

    def __init__(
        self,
        direction: str,
        frame_type: str,
        body: Body,
        offset_ms: int = 0,
    ) -> None: ...

class WsInteraction:
    uri: str
    headers: dict[str, list[str]]
    frames: list[WsFrame]
    recorded_at: str

    def __init__(
        self,
        uri: str,
        headers: dict[str, list[str]] | None = None,
        frames: list[WsFrame] | None = None,
        recorded_at: str | None = None,
    ) -> None: ...

class Cassette:
    version: int
    interactions: list[HttpInteraction]
    played_indices: list[bool]
    grpc_interactions: list[GrpcInteraction]
    grpc_played: list[bool]
    ws_interactions: list[WsInteraction]
    ws_played: list[bool]
    unplayed_count: int

    def __init__(self) -> None: ...
    def add_interaction(self, interaction: HttpInteraction) -> None: ...
    def mark_played(self, index: int) -> None: ...
    def add_grpc_interaction(self, interaction: GrpcInteraction) -> None: ...
    def mark_grpc_played(self, index: int) -> None: ...
    def add_ws_interaction(self, interaction: WsInteraction) -> None: ...
    def mark_ws_played(self, index: int) -> None: ...
    @staticmethod
    def load(path: str) -> Cassette: ...
    def save(self, path: str) -> None: ...
    def __len__(self) -> int: ...

class MatchConfig:
    match_on: list[str]
    ignore_json_paths: list[str]

    def __init__(
        self,
        match_on: list[str] | None = None,
        ignore_json_paths: list[str] | None = None,
    ) -> None: ...

class SecurityConfig:
    filter_headers: list[str]
    filter_query_params: list[str]
    body_scrub_patterns: list[str]
    replacement: str

    def __init__(
        self,
        filter_headers: list[str] | None = None,
        filter_query_params: list[str] | None = None,
        body_scrub_patterns: list[str] | None = None,
        replacement: str | None = None,
    ) -> None: ...

def find_match(
    request: HttpRequest,
    interactions: list[HttpInteraction],
    played: list[bool],
    config: MatchConfig,
) -> tuple[int, HttpInteraction] | None: ...
def find_grpc_match(
    method: str,
    interactions: list[GrpcInteraction],
    played: list[bool],
) -> tuple[int, GrpcInteraction] | None: ...
def find_ws_match(
    uri: str,
    interactions: list[WsInteraction],
    played: list[bool],
) -> tuple[int, WsInteraction] | None: ...
def scrub_interaction(interaction: HttpInteraction, config: SecurityConfig) -> HttpInteraction: ...
def process_body(
    raw_bytes: bytes,
    content_type: str | None = None,
    content_encoding: str | None = None,
) -> Body: ...



from __future__ import annotations

import asyncio
import json
from typing import Any
from unittest.mock import patch

import aiohttp
import aiohttp.client_reqrep
from multidict import CIMultiDict, CIMultiDictProxy
from yarl import URL

from cassetter._core import HttpResponse as _HttpResponse
from cassetter._state import get_current_cassette
from cassetter.cassette import BypassCassette, NoMatchError, RawRequest


class AiohttpInterceptor:
    """Intercepts aiohttp requests by patching ClientSession._request."""

    def __init__(self) -> None:
        self._patcher: Any = None

    def install(self) -> None:
        original_request = aiohttp.ClientSession._request

        async def patched_request(
            session: aiohttp.ClientSession,
            method: str,
            str_or_url: str | URL,
            **kwargs: Any,
        ) -> aiohttp.ClientResponse:
            cassette = get_current_cassette()
            if cassette is None:
                return await original_request(session, method, str_or_url, **kwargs)

            uri = str(URL(str_or_url))

            if cassette.should_bypass(uri):
                return await original_request(session, method, str_or_url, **kwargs)

            headers = _extract_request_headers(kwargs.get("headers"))
            body = _extract_request_body(kwargs)

            hook = cassette.before_record_request
            if hook is not None:
                try:
                    hook(RawRequest(method.upper(), uri, headers, body))
                except BypassCassette:
                    return await original_request(session, method, str_or_url, **kwargs)

            try:
                response = cassette.play(method.upper(), uri, headers, body)
                return _build_aiohttp_response(method, uri, response)
            except NoMatchError:
                if not cassette.can_record:
                    raise

            real_response = await original_request(session, method, str_or_url, **kwargs)
            resp_body = await real_response.read()
            resp_headers = _extract_response_headers(real_response.headers)

            cassette.record(
                method=method.upper(),
                uri=uri,
                request_headers=headers,
                request_body=body,
                status=real_response.status,
                response_headers=resp_headers,
                response_body=resp_body,
            )
            return real_response

        self._patcher = patch.object(aiohttp.ClientSession, "_request", patched_request)
        self._patcher.start()

    def uninstall(self) -> None:
        if self._patcher is not None:
            self._patcher.stop()
            self._patcher = None


def _extract_request_headers(headers: Any) -> dict[str, list[str]]:
    if headers is None:
        return {}
    result: dict[str, list[str]] = {}
    for k, v in headers.items():
        result.setdefault(str(k).lower(), []).append(str(v))
    return result


def _extract_request_body(kwargs: dict[str, Any]) -> bytes | None:
    if "data" in kwargs and kwargs["data"] is not None:
        data = kwargs["data"]
        if isinstance(data, bytes):
            return data
        if isinstance(data, str):
            return data.encode()
    if "json" in kwargs and kwargs["json"] is not None:
        return json.dumps(kwargs["json"]).encode()
    return None


def _extract_response_headers(headers: CIMultiDictProxy[str]) -> dict[str, list[str]]:
    result: dict[str, list[str]] = {}
    for key, value in headers.items():
        result.setdefault(key.lower(), []).append(value)
    return result


def _build_aiohttp_response(method: str, uri: str, response: _HttpResponse) -> aiohttp.ClientResponse:

    body = response.body
    if body.body_type == "json":
        content = json.dumps(body.content).encode()
    elif body.body_type == "text":
        content = body.content.encode() if isinstance(body.content, str) else b""
    elif body.body_type == "binary":
        content = body.content if isinstance(body.content, bytes) else b""
    else:
        content = b""

    headers_multi: CIMultiDict[str] = CIMultiDict()
    for key, values in response.headers.items():
        for v in values:
            headers_multi.add(key, v)

    # Build a mock-like response
    resp = aiohttp.ClientResponse(
        method=method,
        url=URL(uri),
        writer=None,
        continue100=None,
        timer=None,  # type: ignore[arg-type]
        request_info=aiohttp.RequestInfo(
            url=URL(uri),
            method=method,
            headers=CIMultiDictProxy(CIMultiDict()),
            real_url=URL(uri),
        ),
        traces=[],
        loop=asyncio.get_running_loop(),
        session=None,  # type: ignore[arg-type]
    )
    resp.status = response.status
    resp._headers = CIMultiDictProxy(headers_multi)
    resp._body = content
    return resp



from __future__ import annotations

import json
from collections.abc import Awaitable, Callable
from typing import Any

import httpx

from cassetter._core import HttpResponse as _HttpResponse
from cassetter._state import get_current_cassette
from cassetter.cassette import BypassCassette, NoMatchError, RawRequest

_AsyncPassthrough = Callable[[httpx.Request], Awaitable[httpx.Response]]
_SyncPassthrough = Callable[[httpx.Request], httpx.Response]


class _VCRAsyncTransport(httpx.AsyncBaseTransport):
    """Wraps a custom (non-standard) async transport for cassette interception."""

    def __init__(self, real_transport: httpx.AsyncBaseTransport) -> None:
        self._real = real_transport

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        return await _async_intercept(request, lambda r: self._real.handle_async_request(r))


class _VCRSyncTransport(httpx.BaseTransport):
    """Wraps a custom (non-standard) sync transport for cassette interception."""

    def __init__(self, real_transport: httpx.BaseTransport) -> None:
        self._real = real_transport

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        return _sync_intercept(request, lambda r: self._real.handle_request(r))


class HttpxInterceptor:
    """Patches httpx transports to record/replay HTTP interactions."""

    def __init__(self) -> None:
        self._original_async_handle = httpx.AsyncHTTPTransport.handle_async_request
        self._original_sync_handle = httpx.HTTPTransport.handle_request
        self._original_async_init = httpx.AsyncClient.__init__
        self._original_sync_init = httpx.Client.__init__

    def install(self) -> None:
        original_async_handle = self._original_async_handle
        original_sync_handle = self._original_sync_handle
        original_async_init = self._original_async_init
        original_sync_init = self._original_sync_init

        # Patch transport classes to intercept pre-existing clients
        async def patched_async_handle(
            transport_self: httpx.AsyncHTTPTransport, request: httpx.Request
        ) -> httpx.Response:
            return await _async_intercept(request, lambda r: original_async_handle(transport_self, r))

        def patched_sync_handle(transport_self: httpx.HTTPTransport, request: httpx.Request) -> httpx.Response:
            return _sync_intercept(request, lambda r: original_sync_handle(transport_self, r))

        httpx.AsyncHTTPTransport.handle_async_request = patched_async_handle  # type: ignore[assignment,method-assign]
        httpx.HTTPTransport.handle_request = patched_sync_handle  # type: ignore[assignment,method-assign]

        # Patch __init__ to wrap custom (non-standard) transports
        def patched_async_init(client_self: httpx.AsyncClient, **kwargs: Any) -> None:
            original_async_init(client_self, **kwargs)
            if not isinstance(client_self._transport, httpx.AsyncHTTPTransport):
                client_self._transport = _VCRAsyncTransport(client_self._transport)

        def patched_sync_init(client_self: httpx.Client, **kwargs: Any) -> None:
            original_sync_init(client_self, **kwargs)
            if not isinstance(client_self._transport, httpx.HTTPTransport):
                client_self._transport = _VCRSyncTransport(client_self._transport)

        httpx.AsyncClient.__init__ = patched_async_init  # type: ignore[assignment,method-assign]
        httpx.Client.__init__ = patched_sync_init  # type: ignore[assignment,method-assign]

    def uninstall(self) -> None:
        httpx.AsyncHTTPTransport.handle_async_request = self._original_async_handle  # type: ignore[method-assign]
        httpx.HTTPTransport.handle_request = self._original_sync_handle  # type: ignore[method-assign]
        httpx.AsyncClient.__init__ = self._original_async_init  # type: ignore[method-assign]
        httpx.Client.__init__ = self._original_sync_init  # type: ignore[method-assign]


async def _async_intercept(
    request: httpx.Request,
    passthrough: _AsyncPassthrough,
) -> httpx.Response:
    cassette = get_current_cassette()
    if cassette is None:
        return await passthrough(request)

    method = request.method
    uri = str(request.url)

    if cassette.should_bypass(uri):
        return await passthrough(request)

    headers = _extract_headers(request.headers)
    try:
        body = request.content
    except httpx.RequestNotRead:
        body = await request.aread()

    hook = cassette.before_record_request
    if hook is not None:
        try:
            hook(RawRequest(method, uri, headers, body))
        except BypassCassette:
            return await passthrough(request)

    try:
        response = cassette.play(method, uri, headers, body)
        return _build_httpx_response(response, request)
    except NoMatchError:
        if not cassette.can_record:
            raise

    real_response = await passthrough(request)
    await real_response.aread()
    resp_body = real_response.content
    # httpx already decompresses the body, so strip content-encoding
    # to prevent double-decompression in the cassette recorder
    resp_headers = _extract_headers_skip_encoding(real_response.headers)

    cassette.record(
        method=method,
        uri=uri,
        request_headers=headers,
        request_body=body,
        status=real_response.status_code,
        response_headers=resp_headers,
        response_body=resp_body,
    )
    return real_response


def _sync_intercept(
    request: httpx.Request,
    passthrough: _SyncPassthrough,
) -> httpx.Response:
    cassette = get_current_cassette()
    if cassette is None:
        return passthrough(request)

    method = request.method
    uri = str(request.url)

    if cassette.should_bypass(uri):
        return passthrough(request)

    headers = _extract_headers(request.headers)
    body = request.content

    hook = cassette.before_record_request
    if hook is not None:
        try:
            hook(RawRequest(method, uri, headers, body))
        except BypassCassette:
            return passthrough(request)

    try:
        response = cassette.play(method, uri, headers, body)
        return _build_httpx_response(response, request)
    except NoMatchError:
        if not cassette.can_record:
            raise

    real_response = passthrough(request)
    real_response.read()
    resp_body = real_response.content
    # httpx already decompresses the body, so strip content-encoding
    resp_headers = _extract_headers_skip_encoding(real_response.headers)

    cassette.record(
        method=method,
        uri=uri,
        request_headers=headers,
        request_body=body,
        status=real_response.status_code,
        response_headers=resp_headers,
        response_body=resp_body,
    )
    return real_response


def _extract_headers(headers: httpx.Headers) -> dict[str, list[str]]:
    result: dict[str, list[str]] = {}
    for key, value in headers.multi_items():
        result.setdefault(key.lower(), []).append(value)
    return result


def _extract_headers_skip_encoding(headers: httpx.Headers) -> dict[str, list[str]]:
    result: dict[str, list[str]] = {}
    for key, value in headers.multi_items():
        lower = key.lower()
        if lower == "content-encoding":
            continue
        result.setdefault(lower, []).append(value)
    return result


def _build_httpx_response(response: _HttpResponse, request: httpx.Request | None = None) -> httpx.Response:
    headers_list: list[tuple[str, str]] = []
    for key, values in response.headers.items():
        for v in values:
            headers_list.append((key, v))

    body = response.body
    if body.body_type == "json":
        content = json.dumps(body.content).encode()
    elif body.body_type == "text":
        content = body.content.encode() if isinstance(body.content, str) else b""
    elif body.body_type == "binary":
        content = body.content if isinstance(body.content, bytes) else b""
    else:
        content = b""

    return httpx.Response(
        status_code=response.status,
        headers=headers_list,
        content=content,
        request=request,
    )



from __future__ import annotations

import struct
from collections.abc import AsyncIterator
from typing import Any

import grpc
import grpc.aio

from cassetter._core import Body, GrpcResponse
from cassetter._state import get_current_cassette
from cassetter.cassette import NoMatchError


class VCRUnaryUnaryCallable:
    """Wraps a unary-unary gRPC callable for record/replay."""

    def __init__(
        self,
        method: str,
        real_callable: grpc.aio.UnaryUnaryMultiCallable[Any, Any] | None,
        request_serializer: Any,
        response_deserializer: Any,
    ) -> None:
        self._method = method
        self._real = real_callable
        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    async def __call__(
        self,
        request: Any,
        *,
        timeout: float | None = None,
        metadata: Any = None,
        credentials: Any = None,
        wait_for_ready: bool | None = None,
        compression: Any = None,
    ) -> Any:
        cassette = get_current_cassette()
        if cassette is None:
            assert self._real is not None
            return await self._real(
                request,
                timeout=timeout,
                metadata=metadata,
                credentials=credentials,
                wait_for_ready=wait_for_ready,
                compression=compression,
            )

        req_bytes: bytes = self._request_serializer(request)
        req_body = Body("binary", req_bytes)
        md = _metadata_to_dict(metadata)

        try:
            grpc_resp = cassette.play_grpc(self._method)
            content = grpc_resp.body.content if isinstance(grpc_resp.body.content, bytes) else b""
            return self._response_deserializer(content)
        except NoMatchError:
            if not cassette.can_record:
                raise

        assert self._real is not None
        response = await self._real(
            request,
            timeout=timeout,
            metadata=metadata,
            credentials=credentials,
            wait_for_ready=wait_for_ready,
            compression=compression,
        )
        resp_bytes: bytes = response.SerializeToString()
        resp_body = Body("binary", resp_bytes)

        json_debug = _build_json_debug(request, response)

        cassette.record_grpc(
            method=self._method,
            metadata=md,
            request_body=req_body,
            response_body=resp_body,
            json_debug=json_debug,
        )
        return response


class VCRUnaryStreamCallable:
    """Wraps a unary-stream (server streaming) gRPC callable for record/replay."""

    def __init__(
        self,
        method: str,
        real_callable: grpc.aio.UnaryStreamMultiCallable[Any, Any] | None,
        request_serializer: Any,
        response_deserializer: Any,
    ) -> None:
        self._method = method
        self._real = real_callable
        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    def __call__(
        self,
        request: Any,
        *,
        timeout: float | None = None,
        metadata: Any = None,
        credentials: Any = None,
        wait_for_ready: bool | None = None,
        compression: Any = None,
    ) -> AsyncIterator[Any]:
        cassette = get_current_cassette()
        if cassette is None:
            assert self._real is not None
            return self._real(  # type: ignore[no-any-return]
                request,
                timeout=timeout,
                metadata=metadata,
                credentials=credentials,
                wait_for_ready=wait_for_ready,
                compression=compression,
            )

        req_bytes: bytes = self._request_serializer(request)
        req_body = Body("binary", req_bytes)
        md = _metadata_to_dict(metadata)

        try:
            grpc_resp = cassette.play_grpc(self._method)
            return _replay_stream(grpc_resp, self._response_deserializer)
        except NoMatchError:
            if not cassette.can_record:
                raise

        assert self._real is not None
        return self._record_unary_stream(
            request,
            req_body,
            md,
            timeout,
            metadata,
            credentials,
            wait_for_ready,
            compression,
        )

    async def _record_unary_stream(
        self,
        request: Any,
        req_body: Body,
        md: dict[str, list[str]],
        timeout: float | None,
        metadata: Any,
        credentials: Any,
        wait_for_ready: bool | None,
        compression: Any,
    ) -> AsyncIterator[Any]:
        cassette = get_current_cassette()
        assert self._real is not None
        call = self._real(
            request,
            timeout=timeout,
            metadata=metadata,
            credentials=credentials,
            wait_for_ready=wait_for_ready,
            compression=compression,
        )
        chunks: list[bytes] = []
        async for response in call:
            resp_bytes = response.SerializeToString()
            chunks.append(resp_bytes)
            yield response
        resp_body = Body("binary", _encode_chunks(chunks))
        if cassette is not None:
            cassette.record_grpc(
                method=self._method,
                metadata=md,
                request_body=req_body,
                response_body=resp_body,
            )


class VCRStreamUnaryCallable:
    """Wraps a stream-unary (client streaming) gRPC callable for record/replay."""

    def __init__(
        self,
        method: str,
        real_callable: grpc.aio.StreamUnaryMultiCallable[Any, Any] | None,
        request_serializer: Any,
        response_deserializer: Any,
    ) -> None:
        self._method = method
        self._real = real_callable
        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    async def __call__(
        self,
        request_iterator: AsyncIterator[Any],
        *,
        timeout: float | None = None,
        metadata: Any = None,
        credentials: Any = None,
        wait_for_ready: bool | None = None,
        compression: Any = None,
    ) -> Any:
        cassette = get_current_cassette()
        if cassette is None:
            assert self._real is not None
            return await self._real(
                request_iterator,
                timeout=timeout,
                metadata=metadata,
                credentials=credentials,
                wait_for_ready=wait_for_ready,
                compression=compression,
            )

        md = _metadata_to_dict(metadata)
        req_chunks: list[bytes] = []
        async for req in request_iterator:
            req_chunks.append(self._request_serializer(req))
        req_body = Body("binary", _encode_chunks(req_chunks))

        try:
            grpc_resp = cassette.play_grpc(self._method)
            content = grpc_resp.body.content if isinstance(grpc_resp.body.content, bytes) else b""
            return self._response_deserializer(content)
        except NoMatchError:
            if not cassette.can_record:
                raise

        assert self._real is not None
        response = await self._real(
            _iter_bytes(req_chunks, self._response_deserializer),
            timeout=timeout,
            metadata=metadata,
            credentials=credentials,
            wait_for_ready=wait_for_ready,
            compression=compression,
        )
        resp_bytes: bytes = response.SerializeToString()
        resp_body = Body("binary", resp_bytes)

        cassette.record_grpc(
            method=self._method,
            metadata=md,
            request_body=req_body,
            response_body=resp_body,
            json_debug=_build_json_debug(None, response),
        )
        return response


class VCRStreamStreamCallable:
    """Wraps a stream-stream (bidi streaming) gRPC callable for record/replay."""

    def __init__(
        self,
        method: str,
        real_callable: grpc.aio.StreamStreamMultiCallable[Any, Any] | None,
        request_serializer: Any,
        response_deserializer: Any,
    ) -> None:
        self._method = method
        self._real = real_callable
        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    def __call__(
        self,
        request_iterator: AsyncIterator[Any],
        *,
        timeout: float | None = None,
        metadata: Any = None,
        credentials: Any = None,
        wait_for_ready: bool | None = None,
        compression: Any = None,
    ) -> AsyncIterator[Any]:
        cassette = get_current_cassette()
        if cassette is None:
            assert self._real is not None
            return self._real(  # type: ignore[no-any-return]
                request_iterator,
                timeout=timeout,
                metadata=metadata,
                credentials=credentials,
                wait_for_ready=wait_for_ready,
                compression=compression,
            )

        md = _metadata_to_dict(metadata)

        try:
            grpc_resp = cassette.play_grpc(self._method)
            return _replay_stream(grpc_resp, self._response_deserializer)
        except NoMatchError:
            if not cassette.can_record:
                raise

        assert self._real is not None
        return self._record_bidi(request_iterator, md, timeout, metadata, credentials, wait_for_ready, compression)

    async def _record_bidi(
        self,
        request_iterator: AsyncIterator[Any],
        md: dict[str, list[str]],
        timeout: float | None,
        metadata: Any,
        credentials: Any,
        wait_for_ready: bool | None,
        compression: Any,
    ) -> AsyncIterator[Any]:
        cassette = get_current_cassette()
        # Collect all request chunks for recording
        req_chunks: list[bytes] = []
        async for req in request_iterator:
            req_chunks.append(self._request_serializer(req))
        req_body = Body("binary", _encode_chunks(req_chunks))

        assert self._real is not None
        call = self._real(
            _async_iter(req_chunks),
            timeout=timeout,
            metadata=metadata,
            credentials=credentials,
            wait_for_ready=wait_for_ready,
            compression=compression,
        )
        resp_chunks: list[bytes] = []
        async for response in call:
            resp_bytes = response.SerializeToString()
            resp_chunks.append(resp_bytes)
            yield response
        resp_body = Body("binary", _encode_chunks(resp_chunks))
        if cassette is not None:
            cassette.record_grpc(
                method=self._method,
                metadata=md,
                request_body=req_body,
                response_body=resp_body,
            )


class VCRChannel:
    """Wraps a real grpc.aio.Channel to intercept stub method calls."""

    def __init__(self, real_channel: grpc.aio.Channel) -> None:
        self._real = real_channel

    def unary_unary(
        self,
        method: str,
        request_serializer: Any = None,
        response_deserializer: Any = None,
    ) -> VCRUnaryUnaryCallable:
        real_callable = self._real.unary_unary(method, request_serializer, response_deserializer)
        return VCRUnaryUnaryCallable(method, real_callable, request_serializer, response_deserializer)

    def unary_stream(
        self,
        method: str,
        request_serializer: Any = None,
        response_deserializer: Any = None,
    ) -> VCRUnaryStreamCallable:
        real_callable = self._real.unary_stream(method, request_serializer, response_deserializer)
        return VCRUnaryStreamCallable(method, real_callable, request_serializer, response_deserializer)

    def stream_unary(
        self,
        method: str,
        request_serializer: Any = None,
        response_deserializer: Any = None,
    ) -> VCRStreamUnaryCallable:
        real_callable = self._real.stream_unary(method, request_serializer, response_deserializer)
        return VCRStreamUnaryCallable(method, real_callable, request_serializer, response_deserializer)

    def stream_stream(
        self,
        method: str,
        request_serializer: Any = None,
        response_deserializer: Any = None,
    ) -> VCRStreamStreamCallable:
        real_callable = self._real.stream_stream(method, request_serializer, response_deserializer)
        return VCRStreamStreamCallable(method, real_callable, request_serializer, response_deserializer)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._real, name)

    async def close(self) -> None:
        await self._real.close()

    async def __aenter__(self) -> VCRChannel:
        await self._real.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self._real.__aexit__(*args)


class GrpcInterceptor:
    """Patches grpc.aio channel creation to intercept gRPC calls."""

    def __init__(self) -> None:
        self._original_insecure: Any = None
        self._original_secure: Any = None

    def install(self) -> None:
        self._original_insecure = grpc.aio.insecure_channel
        self._original_secure = grpc.aio.secure_channel

        original_insecure = self._original_insecure
        original_secure = self._original_secure

        def patched_insecure(target: str, **kwargs: Any) -> VCRChannel:  # pragma: no cover
            real = original_insecure(target, **kwargs)
            return VCRChannel(real)

        def patched_secure(target: str, credentials: Any, **kwargs: Any) -> VCRChannel:  # pragma: no cover
            real = original_secure(target, credentials, **kwargs)
            return VCRChannel(real)

        grpc.aio.insecure_channel = patched_insecure
        grpc.aio.secure_channel = patched_secure

    def uninstall(self) -> None:
        if self._original_insecure is not None:
            grpc.aio.insecure_channel = self._original_insecure
        if self._original_secure is not None:
            grpc.aio.secure_channel = self._original_secure


# ---------------------------------------------------------------------------
# Chunk encoding (length-prefixed binary)
# ---------------------------------------------------------------------------


def _encode_chunks(chunks: list[bytes]) -> bytes:
    """Concatenate chunks with 4-byte big-endian length prefixes."""
    parts: list[bytes] = []
    for chunk in chunks:
        parts.append(struct.pack(">I", len(chunk)))
        parts.append(chunk)
    return b"".join(parts)


def _decode_chunks(data: bytes) -> list[bytes]:
    """Decode length-prefixed chunks back to a list."""
    chunks: list[bytes] = []
    offset = 0
    while offset < len(data):
        if offset + 4 > len(data):
            break
        (length,) = struct.unpack(">I", data[offset : offset + 4])
        offset += 4
        chunks.append(data[offset : offset + length])
        offset += length
    return chunks


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _metadata_to_dict(metadata: Any) -> dict[str, list[str]]:
    if metadata is None:
        return {}
    result: dict[str, list[str]] = {}
    for key, value in metadata:
        str_val = value if isinstance(value, str) else value.decode("utf-8", errors="replace")
        result.setdefault(key, []).append(str_val)
    return result


async def _replay_stream(grpc_resp: GrpcResponse, deserializer: Any) -> AsyncIterator[Any]:
    body = grpc_resp.body
    data = body.content if isinstance(body.content, bytes) else b""
    chunks = _decode_chunks(data)
    if chunks:
        for chunk in chunks:
            yield deserializer(chunk)
    else:
        # Single message (non-chunked) fallback
        yield deserializer(data)


async def _async_iter(chunks: list[bytes]) -> AsyncIterator[bytes]:
    for chunk in chunks:
        yield chunk


def _iter_bytes(chunks: list[bytes], deserializer: Any) -> AsyncIterator[Any]:
    """Re-create an async iterator of deserialized messages from raw bytes."""
    return _async_iter(chunks)


def _build_json_debug(request: Any, response: Any) -> dict[str, Any] | None:
    try:
        from google.protobuf.json_format import MessageToDict  # pragma: no cover

        result: dict[str, Any] = {}  # pragma: no cover
        if request is not None:  # pragma: no cover
            result["request"] = MessageToDict(request, preserving_proto_field_name=True)
        if response is not None:  # pragma: no cover
            result["response"] = MessageToDict(response, preserving_proto_field_name=True)
        return result or None  # pragma: no cover
    except (ImportError, AttributeError):
        return None



from __future__ import annotations

from typing import Any

from typing_extensions import Protocol, TypedDict

from cassetter.cassette import BeforeRecordRequest


class CassetteConfig(TypedDict, total=False):
    """Configuration for a cassette recording/playback session."""

    record_mode: str
    match_on: list[str]
    ignore_json_paths: list[str]
    filter_headers: list[str]
    filter_query_params: list[str]
    body_scrub_patterns: list[str]
    filter_replacement: str
    cassette_dir: str
    intercept: list[str]
    max_age: str
    on_expiry: str
    ignore_localhost: bool
    ignore_hosts: list[str]
    before_record_request: BeforeRecordRequest


class Interceptor(Protocol):
    """Protocol for HTTP library interceptors."""

    def install(self) -> None: ...
    def uninstall(self) -> None: ...


class AsyncInterceptor(Protocol):
    """Protocol for async HTTP library interceptors."""

    async def install(self) -> None: ...
    async def uninstall(self) -> None: ...


class RequestCallback(Protocol):
    """Called when an interceptor captures an outgoing request."""

    def __call__(
        self,
        method: str,
        uri: str,
        headers: dict[str, list[str]],
        body: bytes | None,
    ) -> Any: ...



from __future__ import annotations

from cassetter._core import (
    Body,
    GrpcInteraction,
    GrpcRequest,
    GrpcResponse,
    HttpInteraction,
    HttpRequest,
    HttpResponse,
    MatchConfig,
    SecurityConfig,
    WsFrame,
    WsInteraction,
)
from cassetter.cassette import (
    BeforeRecordRequest,
    BypassCassette,
    Cassette,
    CassetteExpiredError,
    CassetteExpiredWarning,
    CassetteNotFoundError,
    NoMatchError,
    RawRequest,
)
from cassetter.context import use_cassette
from cassetter.recording import RecordMode

__all__ = [
    "BeforeRecordRequest",
    "Body",
    "BypassCassette",
    "Cassette",
    "CassetteExpiredError",
    "CassetteExpiredWarning",
    "CassetteNotFoundError",
    "GrpcInteraction",
    "GrpcRequest",
    "GrpcResponse",
    "HttpInteraction",
    "HttpRequest",
    "HttpResponse",
    "MatchConfig",
    "NoMatchError",
    "RawRequest",
    "RecordMode",
    "SecurityConfig",
    "WsFrame",
    "WsInteraction",
    "use_cassette",
]



from __future__ import annotations

import contextlib
import os
from collections.abc import Iterator
from typing import Any

from cassetter._core import MatchConfig, SecurityConfig
from cassetter._state import _current_cassette, acquire_patches, release_patches
from cassetter.cassette import BeforeRecordRequest, Cassette
from cassetter.intercept._base import InterceptorProtocol
from cassetter.intercept._httpx import HttpxInterceptor
from cassetter.recording import RecordMode

try:
    from cassetter.intercept._aiohttp import AiohttpInterceptor
except ImportError:  # pragma: no cover
    AiohttpInterceptor = None  # type: ignore[assignment, misc]

try:
    from cassetter.intercept._requests import RequestsInterceptor
except ImportError:  # pragma: no cover
    RequestsInterceptor = None  # type: ignore[assignment, misc]

try:
    from cassetter.intercept._grpc import GrpcInterceptor
except ImportError:  # pragma: no cover
    GrpcInterceptor = None  # type: ignore[assignment, misc]

try:
    from cassetter.intercept._websockets import WebSocketInterceptor
except ImportError:  # pragma: no cover
    WebSocketInterceptor = None  # type: ignore[assignment, misc]

try:
    from cassetter.intercept._urllib3 import Urllib3Interceptor
except ImportError:  # pragma: no cover
    Urllib3Interceptor = None  # type: ignore[assignment, misc]

try:
    from cassetter.intercept._pyreqwest import PyreqwestInterceptor
except ImportError:  # pragma: no cover
    PyreqwestInterceptor = None  # type: ignore[assignment, misc]

_INTERCEPTOR_MAP: dict[str, type[InterceptorProtocol] | None] = {
    "httpx": HttpxInterceptor,
    "aiohttp": AiohttpInterceptor,
    "requests": RequestsInterceptor,
    "grpc": GrpcInterceptor,
    "websockets": WebSocketInterceptor,
    "urllib3": Urllib3Interceptor,
    "pyreqwest": PyreqwestInterceptor,
}

# Interceptors to enable by default, in order. urllib3 subsumes requests,
# so requests is excluded when urllib3 is available.
_AUTO_DETECT_ORDER: list[str] = ["httpx", "urllib3", "aiohttp"]
_SUBSUMED_BY: dict[str, str] = {"requests": "urllib3"}


@contextlib.contextmanager
def use_cassette(
    path: str | os.PathLike[str],
    *,
    record_mode: RecordMode | str = RecordMode.ONCE,
    match_on: list[str] | None = None,
    ignore_json_paths: list[str] | None = None,
    filter_headers: list[str] | None = None,
    filter_query_params: list[str] | None = None,
    body_scrub_patterns: list[str] | None = None,
    filter_replacement: str | None = None,
    intercept: list[str] | None = None,
    max_age: str | None = None,
    on_expiry: str = "warn",
    ignore_localhost: bool = False,
    ignore_hosts: list[str] | None = None,
    before_record_request: BeforeRecordRequest | None = None,
) -> Iterator[Cassette]:
    """Context manager for recording/replaying HTTP interactions.

    Args:
        path: Path to the cassette YAML file.
        record_mode: Controls recording behavior.
        match_on: Fields to match on (default: ["method", "uri"]).
        ignore_json_paths: JSON paths to ignore during matching.
        filter_headers: Headers to filter from cassettes.
        filter_query_params: Query params to filter.
        body_scrub_patterns: Body patterns to scrub.
        filter_replacement: Replacement string for filtered values.
        intercept: HTTP libraries to intercept (default: auto-detect).
    """
    if isinstance(record_mode, str):
        record_mode = RecordMode.from_str(record_mode)

    match_config = MatchConfig(match_on=match_on, ignore_json_paths=ignore_json_paths)

    security_kwargs: dict[str, Any] = {}
    if filter_headers is not None:
        security_kwargs["filter_headers"] = filter_headers
    if filter_query_params is not None:
        security_kwargs["filter_query_params"] = filter_query_params
    if body_scrub_patterns is not None:
        security_kwargs["body_scrub_patterns"] = body_scrub_patterns
    if filter_replacement is not None:
        security_kwargs["replacement"] = filter_replacement
    security_config = SecurityConfig(**security_kwargs)

    cassette = Cassette(
        path,
        record_mode=record_mode,
        match_config=match_config,
        security_config=security_config,
        max_age=max_age,
        on_expiry=on_expiry,
        ignore_localhost=ignore_localhost,
        ignore_hosts=ignore_hosts,
        before_record_request=before_record_request,
    )
    cassette.load()

    interceptor_classes = resolve_interceptors(intercept)

    acquire_patches(interceptor_classes)
    token = _current_cassette.set(cassette)

    try:
        yield cassette
    finally:
        _current_cassette.reset(token)
        release_patches(interceptor_classes)
        cassette.save()


def resolve_interceptors(names: list[str] | None = None) -> list[type[InterceptorProtocol]]:
    """Import and return interceptor classes by name, or auto-detect if None."""
    if names is None:
        return _auto_detect_interceptors()
    interceptors: list[type[InterceptorProtocol]] = []
    for name in names:
        cls = _INTERCEPTOR_MAP.get(name)
        if cls is None:
            if name in _INTERCEPTOR_MAP:
                raise ImportError(f"interceptor {name!r} requires installing the '{name}' extra")
            raise ValueError(f"unknown interceptor: {name!r}")
        interceptors.append(cls)
    return interceptors


def _auto_detect_interceptors() -> list[type[InterceptorProtocol]]:
    """Detect installed HTTP libraries and return interceptor classes for all of them."""
    available: list[str] = []
    for name in _AUTO_DETECT_ORDER:
        if _INTERCEPTOR_MAP.get(name) is not None:
            available.append(name)
    if not available:
        raise ImportError("no HTTP interceptors available - install httpx, urllib3, or aiohttp")
    return [_INTERCEPTOR_MAP[name] for name in available]  # type: ignore[misc]



from __future__ import annotations

import time
from typing import Any

import websockets
import websockets.asyncio.client

from cassetter._core import Body, WsFrame, WsInteraction
from cassetter._state import get_current_cassette
from cassetter.cassette import NoMatchError


class VCRWebSocket:
    """Wraps a real WebSocket connection to record sent/received frames."""

    def __init__(self, real_ws: Any, uri: str, headers: dict[str, list[str]]) -> None:
        self._real = real_ws
        self._uri = uri
        self._headers = headers
        self._frames: list[WsFrame] = []
        self._start_time = time.monotonic()

    async def send(self, message: str | bytes) -> None:
        offset_ms = int((time.monotonic() - self._start_time) * 1000)
        if isinstance(message, bytes):
            body = Body("binary", message)
            frame_type = "binary"
        else:
            body = Body("text", message)
            frame_type = "text"
        self._frames.append(WsFrame("send", frame_type, body, offset_ms))
        await self._real.send(message)

    async def recv(self) -> str | bytes:
        data: str | bytes = await self._real.recv()
        offset_ms = int((time.monotonic() - self._start_time) * 1000)
        if isinstance(data, bytes):
            body = Body("binary", data)
            frame_type = "binary"
        else:
            body = Body("text", data)
            frame_type = "text"
        self._frames.append(WsFrame("recv", frame_type, body, offset_ms))
        return data

    async def close(self, code: int = 1000, reason: str = "") -> None:
        await self._real.close(code, reason)
        self._flush()

    def _flush(self) -> None:
        cassette = get_current_cassette()
        if self._frames and cassette is not None:
            cassette.record_ws(self._uri, self._headers, self._frames)
            self._frames = []

    async def __aenter__(self) -> VCRWebSocket:
        return self

    async def __aexit__(self, *args: Any) -> None:
        self._flush()
        await self._real.close()

    def __aiter__(self) -> VCRWebSocket:
        return self

    async def __anext__(self) -> str | bytes:
        try:
            return await self.recv()
        except websockets.exceptions.ConnectionClosed:
            self._flush()
            raise StopAsyncIteration


class VCRWebSocketReplay:
    """Replays recorded WebSocket frames without a real connection."""

    def __init__(self, interaction: WsInteraction) -> None:
        self._frames = interaction.frames
        self._recv_frames = [f for f in self._frames if f.direction == "recv"]
        self._recv_index = 0

    async def send(self, message: str | bytes) -> None:
        pass

    async def recv(self) -> str | bytes:
        if self._recv_index >= len(self._recv_frames):
            raise StopAsyncIteration
        frame = self._recv_frames[self._recv_index]
        self._recv_index += 1
        return _frame_to_data(frame)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        pass

    async def __aenter__(self) -> VCRWebSocketReplay:
        return self

    async def __aexit__(self, *args: Any) -> None:
        pass

    def __aiter__(self) -> VCRWebSocketReplay:
        return self

    async def __anext__(self) -> str | bytes:
        try:
            return await self.recv()
        except StopAsyncIteration:
            raise


class WebSocketInterceptor:
    """Patches websockets.connect to intercept WebSocket connections."""

    def __init__(self) -> None:
        self._original_connect: Any = None

    def install(self) -> None:
        self._original_connect = websockets.asyncio.client.connect
        original_connect = self._original_connect

        class PatchedConnect:
            def __init__(self, uri: str, **kwargs: Any) -> None:
                self._uri = uri
                self._kwargs = kwargs
                self._ws: VCRWebSocket | VCRWebSocketReplay | None = None

            async def __aenter__(self) -> VCRWebSocket | VCRWebSocketReplay:
                cassette = get_current_cassette()
                if cassette is None:
                    conn = original_connect(self._uri, **self._kwargs)  # pragma: no cover
                    return await conn.__aenter__()  # type: ignore[no-any-return]  # pragma: no cover

                try:
                    interaction = cassette.play_ws(self._uri)
                    self._ws = VCRWebSocketReplay(interaction)
                    return self._ws
                except NoMatchError:
                    if not cassette.can_record:
                        raise

                conn = original_connect(self._uri, **self._kwargs)  # pragma: no cover
                real_ws = await conn.__aenter__()  # pragma: no cover
                headers = _extract_ws_headers(self._kwargs)  # pragma: no cover
                self._ws = VCRWebSocket(real_ws, self._uri, headers)  # pragma: no cover
                return self._ws  # pragma: no cover

            async def __aexit__(self, *args: Any) -> None:
                if self._ws is not None:
                    await self._ws.__aexit__(*args)

        websockets.asyncio.client.connect = PatchedConnect  # type: ignore[assignment,misc]
        websockets.connect = PatchedConnect  # type: ignore[assignment,misc]

    def uninstall(self) -> None:
        if self._original_connect is not None:
            websockets.asyncio.client.connect = self._original_connect  # type: ignore[misc]
            websockets.connect = self._original_connect  # type: ignore[misc]


def _extract_ws_headers(kwargs: dict[str, Any]) -> dict[str, list[str]]:
    extra = kwargs.get("additional_headers") or kwargs.get("extra_headers")
    if extra is None:
        return {}
    result: dict[str, list[str]] = {}
    if isinstance(extra, dict):
        for k, v in extra.items():
            result[k.lower()] = [v] if isinstance(v, str) else list(v)
    return result


def _frame_to_data(frame: WsFrame) -> str | bytes:
    body = frame.body
    if body.body_type == "binary":
        return body.content if isinstance(body.content, bytes) else b""
    if body.body_type == "text":
        return body.content if isinstance(body.content, str) else ""
    return ""



from __future__ import annotations

import json
from typing import Any
from unittest.mock import patch

import requests
import requests.adapters

from cassetter._core import HttpResponse as _HttpResponse
from cassetter._state import get_current_cassette
from cassetter.cassette import BypassCassette, Cassette, NoMatchError, RawRequest


class VCRAdapter(requests.adapters.HTTPAdapter):
    """requests HTTPAdapter that records/replays via a Cassette."""

    def __init__(self, cassette: Cassette, real_adapter: requests.adapters.HTTPAdapter) -> None:
        super().__init__()
        self._cassette = cassette
        self._real_adapter = real_adapter

    def send(  # type: ignore[override]
        self,
        request: requests.PreparedRequest,
        stream: bool = False,
        timeout: float | tuple[float, float] | None = None,
        verify: bool | str = True,
        cert: str | tuple[str, str] | None = None,
        proxies: dict[str, str] | None = None,
    ) -> requests.Response:
        method = (request.method or "GET").upper()
        uri = request.url or ""
        headers = _extract_headers(request.headers)
        body = request.body if isinstance(request.body, bytes) else (request.body.encode() if request.body else None)

        try:
            response = self._cassette.play(method, uri, headers, body)
            return _build_requests_response(request, response)
        except NoMatchError:
            if not self._cassette.can_record:
                raise

        real_response = self._real_adapter.send(
            request, stream=stream, timeout=timeout, verify=verify, cert=cert, proxies=proxies
        )
        resp_headers = _extract_headers(real_response.headers)

        self._cassette.record(
            method=method,
            uri=uri,
            request_headers=headers,
            request_body=body,
            status=real_response.status_code,
            response_headers=resp_headers,
            response_body=real_response.content,
        )
        return real_response


class RequestsInterceptor:
    """Intercepts requests by patching Session.send."""

    def __init__(self) -> None:
        self._patcher: Any = None

    def install(self) -> None:
        original_send = requests.Session.send

        def patched_send(
            session: requests.Session, request: requests.PreparedRequest, **kwargs: Any
        ) -> requests.Response:
            cassette = get_current_cassette()
            if cassette is None:
                return original_send(session, request, **kwargs)

            uri = request.url or ""

            if cassette.should_bypass(uri):
                return original_send(session, request, **kwargs)

            method = (request.method or "GET").upper()
            headers = _extract_headers(request.headers)
            raw_body = request.body
            body = raw_body if isinstance(raw_body, bytes) else (raw_body.encode() if raw_body else None)

            hook = cassette.before_record_request
            if hook is not None:
                try:
                    hook(RawRequest(method, uri, headers, body))
                except BypassCassette:
                    return original_send(session, request, **kwargs)

            try:
                response = cassette.play(method, uri, headers, body)
                return _build_requests_response(request, response)
            except NoMatchError:
                if not cassette.can_record:
                    raise

            real_response = original_send(session, request, **kwargs)
            resp_headers = _extract_headers(real_response.headers)

            cassette.record(
                method=method,
                uri=uri,
                request_headers=headers,
                request_body=body,
                status=real_response.status_code,
                response_headers=resp_headers,
                response_body=real_response.content,
            )
            return real_response

        self._patcher = patch.object(requests.Session, "send", patched_send)
        self._patcher.start()

    def uninstall(self) -> None:
        if self._patcher is not None:
            self._patcher.stop()
            self._patcher = None


def _extract_headers(headers: Any) -> dict[str, list[str]]:
    result: dict[str, list[str]] = {}
    if headers is None:
        return result
    for key, value in headers.items():
        result.setdefault(str(key).lower(), []).append(str(value))
    return result


def _build_requests_response(request: requests.PreparedRequest, response: _HttpResponse) -> requests.Response:

    body = response.body
    if body.body_type == "json":
        content = json.dumps(body.content).encode()
    elif body.body_type == "text":
        content = body.content.encode() if isinstance(body.content, str) else b""
    elif body.body_type == "binary":
        content = body.content if isinstance(body.content, bytes) else b""
    else:
        content = b""

    resp = requests.Response()
    resp.status_code = response.status
    resp._content = content
    resp.encoding = "utf-8"
    resp.url = request.url or ""
    resp.request = request

    for key, values in response.headers.items():
        for v in values:
            resp.headers[key] = v

    return resp



from __future__ import annotations

import os
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import pytest

from cassetter._core import MatchConfig, SecurityConfig
from cassetter._state import _current_cassette, acquire_patches, release_patches
from cassetter._types import CassetteConfig
from cassetter.cassette import Cassette
from cassetter.context import resolve_interceptors
from cassetter.intercept._base import InterceptorProtocol
from cassetter.recording import RecordMode


@pytest.fixture(scope="module")
def vcr_config() -> CassetteConfig:
    """Override this fixture to provide default VCR configuration."""
    return CassetteConfig()


def _resolve_cassette(
    node_name: str,
    marker_args: tuple[Any, ...],
    marker_kwargs: dict[str, Any],
    vcr_config: CassetteConfig,
    cli_record_mode: str | None,
    test_fspath: str,
) -> tuple[Cassette, list[type[InterceptorProtocol]]]:
    """Resolve cassette configuration and create a Cassette instance."""
    cassette_name = node_name + ".yaml"
    cassette_dir = vcr_config.get("cassette_dir", "cassettes")
    record_mode_str = vcr_config.get("record_mode", "none")

    # Marker can override
    if marker_args:
        cassette_name = marker_args[0]
    if "record_mode" in marker_kwargs:
        record_mode_str = marker_kwargs["record_mode"]
    if "cassette_dir" in marker_kwargs:
        cassette_dir = marker_kwargs["cassette_dir"]

    # CLI override
    if cli_record_mode is not None:
        record_mode_str = cli_record_mode

    record_mode = RecordMode.from_str(record_mode_str)

    # Resolve cassette path: {test_dir}/cassettes/{test_file_stem}/{cassette_name}
    test_file = Path(test_fspath)
    test_dir = str(test_file.parent)
    test_file_stem = test_file.stem
    cassette_path = os.path.join(test_dir, cassette_dir, test_file_stem, cassette_name)

    match_config = MatchConfig(
        match_on=vcr_config.get("match_on"),
        ignore_json_paths=vcr_config.get("ignore_json_paths"),
    )

    security_kwargs: dict[str, Any] = {}
    if "filter_headers" in vcr_config:
        security_kwargs["filter_headers"] = vcr_config["filter_headers"]
    if "filter_query_params" in vcr_config:
        security_kwargs["filter_query_params"] = vcr_config["filter_query_params"]
    if "body_scrub_patterns" in vcr_config:
        security_kwargs["body_scrub_patterns"] = vcr_config["body_scrub_patterns"]
    if "filter_replacement" in vcr_config:
        security_kwargs["replacement"] = vcr_config["filter_replacement"]
    security_config = SecurityConfig(**security_kwargs)

    max_age = marker_kwargs.get("max_age", vcr_config.get("max_age"))
    on_expiry = marker_kwargs.get("on_expiry", vcr_config.get("on_expiry", "warn"))

    ignore_localhost = vcr_config.get("ignore_localhost", False)
    ignore_hosts = vcr_config.get("ignore_hosts")
    before_record_request = vcr_config.get("before_record_request")

    cassette = Cassette(
        cassette_path,
        record_mode=record_mode,
        match_config=match_config,
        security_config=security_config,
        max_age=max_age,
        on_expiry=on_expiry,
        ignore_localhost=ignore_localhost,
        ignore_hosts=ignore_hosts,
        before_record_request=before_record_request,
    )
    cassette.load()

    intercept_names = vcr_config.get("intercept")
    interceptor_classes = resolve_interceptors(intercept_names)
    return cassette, interceptor_classes


@pytest.fixture(autouse=True)
def cassette(request: pytest.FixtureRequest, vcr_config: CassetteConfig) -> Iterator[Cassette | None]:
    """Activates cassette recording/replay for tests marked with @pytest.mark.vcr."""
    marker = request.node.get_closest_marker("vcr")
    if marker is None:
        yield None
        return

    cli_record_mode = request.config.getoption("--record-mode", default=None)

    # Include class name for class-based tests (e.g. TestOpenAI.test_query)
    cls = request.node.cls
    node_name = f"{cls.__name__}.{request.node.name}" if cls else request.node.name

    cassette, interceptor_classes = _resolve_cassette(
        node_name=node_name,
        marker_args=marker.args,
        marker_kwargs=dict(marker.kwargs),
        vcr_config=vcr_config,
        cli_record_mode=cli_record_mode,
        test_fspath=str(request.path),
    )

    # Track loaded cassette paths for orphan detection
    _loaded_cassettes = getattr(request.config, "_vcr_loaded_cassettes", None)
    if _loaded_cassettes is not None:
        _loaded_cassettes.add(os.path.abspath(cassette.path))

    acquire_patches(interceptor_classes)
    token = _current_cassette.set(cassette)

    yield cassette

    # Reset context and release patches
    _current_cassette.reset(token)
    release_patches(interceptor_classes)
    cassette.save()



from __future__ import annotations

from cassetter.pytest_plugin.fixtures import cassette as cassette, vcr_config as vcr_config
from cassetter.pytest_plugin.markers import configure as configure
from cassetter.pytest_plugin.orphans import add_options, check_orphans as check_orphans, session_finish

__all__ = ["cassette", "vcr_config", "configure", "check_orphans"]


def pytest_configure(config: object) -> None:
    configure(config)


def pytest_addoption(parser: object) -> None:
    add_options(parser)


def pytest_sessionfinish(session: object, exitstatus: int) -> None:
    session_finish(session)



from __future__ import annotations

from typing import Any


def configure(config: Any) -> None:
    """Register the @pytest.mark.vcr marker."""
    config.addinivalue_line(
        "markers",
        "vcr(cassette_name, **kwargs): Mark test to use VCR cassette recording/replay.",
    )

    # Initialize cassette tracking set
    try:
        config.option
    except AttributeError:
        return

    config._vcr_loaded_cassettes = set()



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Any
from unittest.mock import patch

import pyreqwest_impersonate as pri

from cassetter._core import HttpResponse as _HttpResponse
from cassetter._state import get_current_cassette
from cassetter.cassette import BypassCassette, NoMatchError, RawRequest

_METHODS_WITH_BODY = frozenset({"post", "put", "patch"})
_ALL_METHODS = ("get", "head", "options", "delete", "post", "put", "patch", "request")


@dataclass
class _ReplayResponse:
    """Lightweight stand-in for pyreqwest_impersonate's native Response.

    The native ``Response`` cannot be constructed from Python, so we mimic its
    public interface for cassette playback.
    """

    status_code: int
    headers: dict[str, str]
    content: bytes
    url: str
    cookies: dict[str, str] = field(default_factory=dict)
    encoding: str = "UTF-8"

    @property
    def text(self) -> str:
        return self.content.decode(self.encoding, errors="replace")

    @property
    def text_plain(self) -> str:
        return self.text

    @property
    def text_markdown(self) -> str:
        return self.text

    def json(self) -> Any:
        return json.loads(self.content)


class PyreqwestInterceptor:
    """Intercepts pyreqwest_impersonate by patching every Client HTTP method."""

    def __init__(self) -> None:
        self._patchers: list[Any] = []

    def install(self) -> None:
        for method_name in _ALL_METHODS:
            original = getattr(pri.Client, method_name)
            patcher = patch.object(
                pri.Client,
                method_name,
                _make_wrapper(method_name, original),
            )
            patcher.start()
            self._patchers.append(patcher)

    def uninstall(self) -> None:
        for patcher in reversed(self._patchers):
            patcher.stop()
        self._patchers.clear()


def _make_wrapper(
    method_name: str,
    original: Any,
) -> Any:
    if method_name == "request":

        def request_wrapper(client: Any, method: str, url: str, **kwargs: Any) -> Any:
            return _intercept(original, client, method.upper(), url, (method, url), kwargs)

        return request_wrapper

    http_method = method_name.upper()

    def method_wrapper(client: Any, url: str, **kwargs: Any) -> Any:
        return _intercept(original, client, http_method, url, (url,), kwargs)

    return method_wrapper


def _intercept(
    original: Any,
    client: Any,
    method: str,
    url: str,
    original_args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    cassette = get_current_cassette()
    if cassette is None:
        return original(client, *original_args, **kwargs)

    if cassette.should_bypass(url):
        return original(client, *original_args, **kwargs)

    norm_headers = _extract_headers(kwargs.get("headers"))
    body = _extract_body(
        content=kwargs.get("content"),
        data=kwargs.get("data"),
        json_payload=kwargs.get("json"),
    )

    hook = cassette.before_record_request
    if hook is not None:
        try:
            hook(RawRequest(method, url, norm_headers, body))
        except BypassCassette:
            return original(client, *original_args, **kwargs)

    try:
        response = cassette.play(method, url, norm_headers, body)
        return _build_replay_response(url, response)
    except NoMatchError:
        if not cassette.can_record:
            raise

    real_response = original(client, *original_args, **kwargs)
    resp_headers = _extract_headers(real_response.headers)

    cassette.record(
        method=method,
        uri=real_response.url,
        request_headers=norm_headers,
        request_body=body,
        status=real_response.status_code,
        response_headers=resp_headers,
        response_body=real_response.content,
    )
    return real_response


def _extract_headers(headers: dict[str, str] | None) -> dict[str, list[str]]:
    if headers is None:
        return {}
    return {k.lower(): [v] for k, v in headers.items()}


def _extract_body(
    *,
    content: bytes | None,
    data: dict[str, str] | str | None,
    json_payload: Any,
) -> bytes | None:
    if content is not None:
        return content
    if data is not None:
        if isinstance(data, bytes):
            return data
        if isinstance(data, str):
            return data.encode()
    if json_payload is not None:
        return json.dumps(json_payload).encode()
    return None


def _build_replay_response(url: str, response: _HttpResponse) -> _ReplayResponse:
    body = response.body
    if body.body_type == "json":
        content = json.dumps(body.content).encode()
    elif body.body_type == "text":
        content = body.content.encode() if isinstance(body.content, str) else b""
    elif body.body_type == "binary":
        content = body.content if isinstance(body.content, bytes) else b""
    else:
        content = b""

    flat_headers: dict[str, str] = {}
    for key, values in response.headers.items():
        flat_headers[key] = values[-1] if values else ""

    return _ReplayResponse(
        status_code=response.status,
        headers=flat_headers,
        content=content,
        url=url,
    )



from __future__ import annotations

import os
import warnings
from typing import Any


def add_options(parser: Any) -> None:
    """Add --vcr-check-orphans CLI option."""
    group = parser.getgroup("vcr", "VCR cassette recording")
    group.addoption(
        "--record-mode",
        action="store",
        default=None,
        choices=["none", "new_episodes", "all", "once"],
        help="VCR record mode: none, new_episodes, all, once",
    )
    group.addoption(
        "--vcr-check-orphans",
        action="store",
        default=None,
        metavar="DIR",
        help="Check for orphaned cassette files in DIR that were not loaded during the test run.",
    )


def session_finish(session: Any) -> None:
    """Check for orphaned cassettes at the end of the test session."""
    config = session.config
    orphan_dir = config.getoption("--vcr-check-orphans", default=None)
    if orphan_dir is None:
        return

    loaded: set[str] = getattr(config, "_vcr_loaded_cassettes", set())
    orphan_dir = os.path.abspath(orphan_dir)

    orphans: list[str] = []
    for root, _dirs, files in os.walk(orphan_dir):
        for f in files:
            if f.endswith((".yaml", ".yml")):
                full_path = os.path.abspath(os.path.join(root, f))
                if full_path not in loaded:
                    orphans.append(os.path.relpath(full_path, orphan_dir))

    if orphans:
        orphan_list = "\n  ".join(sorted(orphans))
        warnings.warn(
            f"Found {len(orphans)} orphaned cassette file(s) in {orphan_dir}:\n  {orphan_list}",
            stacklevel=1,
        )


def check_orphans(cassette_dir: str, loaded_paths: set[str]) -> list[str]:
    """Return list of cassette files in cassette_dir not in loaded_paths."""
    orphans = []
    cassette_dir = os.path.abspath(cassette_dir)
    for root, _dirs, files in os.walk(cassette_dir):
        for f in files:
            if f.endswith((".yaml", ".yml")):
                full_path = os.path.abspath(os.path.join(root, f))
                if full_path not in loaded_paths:
                    orphans.append(os.path.relpath(full_path, cassette_dir))
    return sorted(orphans)



from __future__ import annotations

import io
import json
from typing import Any
from unittest.mock import patch

import urllib3
import urllib3.connectionpool
import urllib3.response

from cassetter._core import HttpResponse as _HttpResponse
from cassetter._state import get_current_cassette
from cassetter.cassette import BypassCassette, NoMatchError, RawRequest


class Urllib3Interceptor:
    """Intercepts HTTP traffic by patching urllib3.HTTPConnectionPool.urlopen."""

    def __init__(self) -> None:
        self._patcher: Any = None

    def install(self) -> None:
        original_urlopen = urllib3.connectionpool.HTTPConnectionPool.urlopen

        def patched_urlopen(
            pool: urllib3.connectionpool.HTTPConnectionPool,
            method: str,
            url: str,
            body: bytes | str | None = None,
            headers: Any = None,
            **kwargs: Any,
        ) -> urllib3.response.HTTPResponse:
            cassette = get_current_cassette()
            if cassette is None:
                return original_urlopen(pool, method, url, body=body, headers=headers, **kwargs)  # type: ignore[return-value]

            full_url = _reconstruct_url(pool, url)

            if cassette.should_bypass(full_url):
                return original_urlopen(pool, method, url, body=body, headers=headers, **kwargs)  # type: ignore[return-value]

            norm_method = method.upper()
            norm_headers = _extract_headers(headers)
            norm_body = body.encode() if isinstance(body, str) else body

            hook = cassette.before_record_request
            if hook is not None:
                try:
                    hook(RawRequest(norm_method, full_url, norm_headers, norm_body))
                except BypassCassette:
                    return original_urlopen(pool, method, url, body=body, headers=headers, **kwargs)  # type: ignore[return-value]

            try:
                response = cassette.play(norm_method, full_url, norm_headers, norm_body)
                return _build_urllib3_response(response, full_url)
            except NoMatchError:
                if not cassette.can_record:
                    raise

            real_response = original_urlopen(pool, method, url, body=body, headers=headers, **kwargs)
            resp_body = real_response.data
            resp_headers = _extract_headers(real_response.headers)

            cassette.record(
                method=norm_method,
                uri=full_url,
                request_headers=norm_headers,
                request_body=norm_body,
                status=real_response.status,
                response_headers=resp_headers,
                response_body=resp_body,
            )

            record_headers = urllib3._collections.HTTPHeaderDict(real_response.headers)
            if "content-length" in record_headers:
                record_headers["content-length"] = str(len(resp_body))

            return urllib3.response.HTTPResponse(
                body=io.BytesIO(resp_body),
                headers=record_headers,
                status=real_response.status,
                preload_content=False,
                decode_content=False,
                request_url=full_url,
            )

        self._patcher = patch.object(urllib3.connectionpool.HTTPConnectionPool, "urlopen", patched_urlopen)
        self._patcher.start()

    def uninstall(self) -> None:
        if self._patcher is not None:
            self._patcher.stop()
            self._patcher = None


def _reconstruct_url(pool: urllib3.connectionpool.HTTPConnectionPool, path: str) -> str:
    scheme = pool.scheme
    host = pool.host
    port = pool.port
    if port and not _is_default_port(scheme, port):
        host_part = f"{host}:{port}"
    else:
        host_part = host
    return f"{scheme}://{host_part}{path}"


def _is_default_port(scheme: str, port: int) -> bool:
    return (scheme == "http" and port == 80) or (scheme == "https" and port == 443)


def _extract_headers(headers: Any) -> dict[str, list[str]]:
    result: dict[str, list[str]] = {}
    if headers is None:
        return result
    for key, value in headers.items():
        result.setdefault(str(key).lower(), []).append(str(value))
    return result


def _build_urllib3_response(response: _HttpResponse, request_url: str) -> urllib3.response.HTTPResponse:
    body = response.body
    if body.body_type == "json":
        content = json.dumps(body.content).encode()
    elif body.body_type == "text":
        content = body.content.encode() if isinstance(body.content, str) else b""
    elif body.body_type == "binary":
        content = body.content if isinstance(body.content, bytes) else b""
    else:
        content = b""

    resp_headers = urllib3._collections.HTTPHeaderDict()
    for key, values in response.headers.items():
        for v in values:
            resp_headers.add(key, v)

    if "content-length" in resp_headers:
        resp_headers["content-length"] = str(len(content))

    return urllib3.response.HTTPResponse(
        body=io.BytesIO(content),
        headers=resp_headers,
        status=response.status,
        preload_content=False,
        decode_content=False,
        request_url=request_url,
    )



from __future__ import annotations

import enum


class RecordMode(enum.Enum):
    """Controls how cassette recording/playback behaves."""

    NONE = enum.auto()
    """Never record - only replay from existing cassettes. Raises if no match found."""

    NEW_EPISODES = enum.auto()
    """Replay existing interactions. Record new ones that don't match."""

    ALL = enum.auto()
    """Record all interactions, overwriting any existing cassette."""

    ONCE = enum.auto()
    """Record if cassette doesn't exist. Replay if it does."""

    @classmethod
    def from_str(cls, value: str) -> RecordMode:
        mapping = {
            "none": cls.NONE,
            "new_episodes": cls.NEW_EPISODES,
            "all": cls.ALL,
            "once": cls.ONCE,
        }
        normalized = value.lower().replace("-", "_")
        if normalized not in mapping:
            raise ValueError(f"unknown record mode: {value!r}, expected one of {list(mapping)}")
        return mapping[normalized]



from __future__ import annotations

import fnmatch
import os
import re
import warnings
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from urllib.parse import urlparse

from cassetter._core import (
    Body,
    Cassette as _RustCassette,
    GrpcInteraction,
    GrpcRequest,
    GrpcResponse,
    HttpInteraction,
    HttpRequest,
    HttpResponse,
    MatchConfig,
    SecurityConfig,
    WsFrame,
    WsInteraction,
    find_grpc_match,
    find_match,
    find_ws_match,
    process_body,
    scrub_interaction,
)
from cassetter.recording import RecordMode


class CassetteNotFoundError(Exception):
    """Raised when a cassette file is not found and record mode doesn't allow recording."""


class CassetteExpiredWarning(UserWarning):
    """Emitted when a cassette is older than the configured max_age."""


class CassetteExpiredError(Exception):
    """Raised when a cassette is older than the configured max_age and on_expiry is 'fail'."""


class NoMatchError(Exception):
    """Raised when no matching interaction is found in the cassette."""


class BypassCassette(Exception):
    """Raised from a before_record_request hook to let the request pass through without recording."""


@dataclass(slots=True)
class RawRequest:
    """Raw HTTP request data passed to the before_record_request hook."""

    method: str
    uri: str
    headers: dict[str, list[str]]
    body: bytes | None


BeforeRecordRequest = Callable[[RawRequest], None]


_DURATION_RE = re.compile(r"^(\d+)([dhw])$")
_UNIT_MAP = {"d": "days", "h": "hours", "w": "weeks"}


def _parse_duration(s: str) -> timedelta:
    """Parse a duration string like '30d', '24h', '4w' into a timedelta."""
    m = _DURATION_RE.match(s)
    if m is None:
        raise ValueError(f"invalid duration string: {s!r} (expected <number><d|h|w>)")
    value, unit = int(m.group(1)), m.group(2)
    return timedelta(**{_UNIT_MAP[unit]: value})


class Cassette:
    """Python wrapper around the Rust Cassette providing record/replay logic."""

    def __init__(
        self,
        path: str | os.PathLike[str],
        *,
        record_mode: RecordMode = RecordMode.ONCE,
        match_config: MatchConfig | None = None,
        security_config: SecurityConfig | None = None,
        max_age: str | None = None,
        on_expiry: str = "warn",
        ignore_localhost: bool = False,
        ignore_hosts: list[str] | None = None,
        before_record_request: BeforeRecordRequest | None = None,
    ) -> None:
        self._path = os.fspath(path)
        self._record_mode = record_mode
        self._match_config = match_config or MatchConfig()
        self._security_config = security_config or SecurityConfig()
        self._max_age = _parse_duration(max_age) if max_age is not None else None
        self._on_expiry = on_expiry
        self._ignore_localhost = ignore_localhost
        self._ignore_hosts = ignore_hosts or []
        self._before_record_request = before_record_request
        self._inner: _RustCassette | None = None
        self._dirty = False

    @property
    def path(self) -> str:
        return self._path

    @property
    def record_mode(self) -> RecordMode:
        return self._record_mode

    @property
    def ignore_localhost(self) -> bool:
        return self._ignore_localhost

    @property
    def before_record_request(self) -> BeforeRecordRequest | None:
        return self._before_record_request

    def should_bypass(self, uri: str) -> bool:
        """Check if a request URI should bypass the cassette entirely."""
        if self._ignore_localhost and _is_localhost(uri):
            return True
        if self._ignore_hosts:
            host = urlparse(uri).hostname or ""
            for pattern in self._ignore_hosts:
                if fnmatch.fnmatch(host, pattern):
                    return True
        return False

    @property
    def interactions(self) -> list[HttpInteraction]:
        if self._inner is None:
            return []
        return self._inner.interactions

    @property
    def grpc_interactions(self) -> list[GrpcInteraction]:
        if self._inner is None:
            return []
        return self._inner.grpc_interactions

    @property
    def ws_interactions(self) -> list[WsInteraction]:
        if self._inner is None:
            return []
        return self._inner.ws_interactions

    def load(self) -> None:
        """Load the cassette from disk, or create a new one based on record mode."""
        exists = os.path.exists(self._path)

        if self._record_mode == RecordMode.ALL or not exists:
            self._inner = _RustCassette()
            if self._record_mode == RecordMode.ALL:
                self._dirty = True
            return

        self._inner = _RustCassette.load(self._path)
        self._check_expiry()

    def _check_expiry(self) -> None:
        """Check if the cassette is expired based on max_age, and apply on_expiry action."""
        if self._max_age is None or self._inner is None:
            return

        newest = self._newest_recorded_at()
        if newest is None:
            return

        cutoff = datetime.now(timezone.utc) - self._max_age
        if newest >= cutoff:
            return

        age_days = (datetime.now(timezone.utc) - newest).days
        msg = f"cassette {self._path!r} is {age_days} days old (max_age={self._max_age})"

        if self._on_expiry == "fail":
            raise CassetteExpiredError(msg)
        if self._on_expiry == "rerecord":
            self._inner = _RustCassette()
            self._dirty = True
            return
        warnings.warn(msg, CassetteExpiredWarning, stacklevel=3)

    def _newest_recorded_at(self) -> datetime | None:
        """Return the newest recorded_at timestamp across all interaction types."""
        assert self._inner is not None  # caller guards this
        timestamps: list[str] = []
        for http_i in self._inner.interactions:
            timestamps.append(http_i.recorded_at)
        for grpc_i in self._inner.grpc_interactions:
            timestamps.append(grpc_i.recorded_at)
        for ws_i in self._inner.ws_interactions:
            timestamps.append(ws_i.recorded_at)
        if not timestamps:
            return None
        return max(datetime.fromisoformat(ts.replace("Z", "+00:00")) for ts in timestamps)

    def save(self) -> None:
        """Save the cassette to disk if modified and has any interactions."""
        if self._inner is not None and self._dirty and len(self._inner) > 0:
            self._inner.save(self._path)
            self._dirty = False

    @property
    def can_record(self) -> bool:
        return self._record_mode in (RecordMode.ALL, RecordMode.NEW_EPISODES, RecordMode.ONCE)

    def play(
        self,
        method: str,
        uri: str,
        headers: dict[str, list[str]],
        body: bytes | None,
    ) -> HttpResponse:
        """Find a matching response for the given request, or raise NoMatchError."""
        if self._inner is None:
            raise NoMatchError("cassette not loaded")

        content_type = _get_header(headers, "content-type")
        content_encoding = _get_header(headers, "content-encoding")
        processed_body = process_body(body or b"", content_type, content_encoding)

        request = HttpRequest(method, uri, headers, processed_body)
        result = find_match(request, self._inner.interactions, self._inner.played_indices, self._match_config)

        if result is None:
            raise NoMatchError(f"no matching interaction for {method} {uri}")

        idx, interaction = result
        self._inner.mark_played(idx)
        return interaction.response

    def record(
        self,
        method: str,
        uri: str,
        request_headers: dict[str, list[str]],
        request_body: bytes | None,
        status: int,
        response_headers: dict[str, list[str]],
        response_body: bytes | None,
    ) -> HttpResponse:
        """Record an interaction and return the response."""
        req_ct = _get_header(request_headers, "content-type")
        req_ce = _get_header(request_headers, "content-encoding")
        req_body = process_body(request_body or b"", req_ct, req_ce)

        resp_ct = _get_header(response_headers, "content-type")
        resp_ce = _get_header(response_headers, "content-encoding")
        resp_body = process_body(response_body or b"", resp_ct, resp_ce)

        # Remove content-encoding since we've already decompressed
        clean_resp_headers = {k: v for k, v in response_headers.items() if k.lower() != "content-encoding"}

        request = HttpRequest(method, uri, request_headers, req_body)
        response = HttpResponse(status, clean_resp_headers, resp_body)
        recorded_at = datetime.now(timezone.utc).isoformat()
        interaction = HttpInteraction(request, response, recorded_at)

        # Apply security filtering
        interaction = scrub_interaction(interaction, self._security_config)

        if self._inner is None:
            self._inner = _RustCassette()

        self._inner.add_interaction(interaction)
        self._dirty = True
        return interaction.response

    # --- gRPC ---

    def play_grpc(self, method: str) -> GrpcResponse:
        """Find a matching gRPC response for the given method, or raise NoMatchError."""
        if self._inner is None:
            raise NoMatchError("cassette not loaded")

        result = find_grpc_match(method, self._inner.grpc_interactions, self._inner.grpc_played)
        if result is None:
            raise NoMatchError(f"no matching gRPC interaction for {method}")

        idx, interaction = result
        self._inner.mark_grpc_played(idx)
        return interaction.response

    def record_grpc(
        self,
        method: str,
        metadata: dict[str, list[str]],
        request_body: Body,
        response_body: Body,
        status_code: int = 0,
        status_message: str = "OK",
        response_metadata: dict[str, list[str]] | None = None,
        json_debug: dict[str, object] | None = None,
    ) -> GrpcResponse:
        """Record a gRPC interaction and return the response."""
        request = GrpcRequest(method, metadata, request_body)
        response = GrpcResponse(status_code, status_message, response_metadata, response_body)
        recorded_at = datetime.now(timezone.utc).isoformat()
        interaction = GrpcInteraction(request, response, recorded_at, json_debug)

        if self._inner is None:
            self._inner = _RustCassette()

        self._inner.add_grpc_interaction(interaction)
        self._dirty = True
        return interaction.response

    # --- WebSocket ---

    def play_ws(self, uri: str) -> WsInteraction:
        """Find a matching WebSocket interaction for the given URI, or raise NoMatchError."""
        if self._inner is None:
            raise NoMatchError("cassette not loaded")

        result = find_ws_match(uri, self._inner.ws_interactions, self._inner.ws_played)
        if result is None:
            raise NoMatchError(f"no matching WebSocket interaction for {uri}")

        idx, interaction = result
        self._inner.mark_ws_played(idx)
        return interaction

    def record_ws(
        self,
        uri: str,
        headers: dict[str, list[str]],
        frames: list[WsFrame],
    ) -> None:
        """Record a WebSocket interaction."""
        recorded_at = datetime.now(timezone.utc).isoformat()
        interaction = WsInteraction(uri, headers, frames, recorded_at)

        if self._inner is None:
            self._inner = _RustCassette()

        self._inner.add_ws_interaction(interaction)
        self._dirty = True


_LOCALHOST_HOSTS = frozenset({"localhost", "127.0.0.1", "[::1]", "::1"})


def _is_localhost(uri: str) -> bool:
    host = urlparse(uri).hostname or ""
    return host in _LOCALHOST_HOSTS


def _get_header(headers: dict[str, list[str]], name: str) -> str | None:
    """Case-insensitive header lookup, returns first value or None."""
    name_lower = name.lower()
    for key, values in headers.items():
        if key.lower() == name_lower and values:
            return values[0]
    return None



from __future__ import annotations

from urllib.parse import urlparse

from typing_extensions import Protocol

_LOCALHOST_HOSTS = frozenset({"localhost", "127.0.0.1", "[::1]", "::1"})


def is_localhost(uri: str) -> bool:
    host = urlparse(uri).hostname or ""
    return host in _LOCALHOST_HOSTS


class InterceptorProtocol(Protocol):
    """Protocol that all interceptors must satisfy."""

    def install(self) -> None: ...
    def uninstall(self) -> None: ...



from __future__ import annotations

import threading
from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cassetter.cassette import Cassette
    from cassetter.intercept._base import InterceptorProtocol

_current_cassette: ContextVar[Cassette | None] = ContextVar("_current_cassette", default=None)

_lock = threading.Lock()
_installed: dict[type, tuple[InterceptorProtocol, int]] = {}


def get_current_cassette() -> Cassette | None:
    return _current_cassette.get()


def acquire_patches(interceptor_classes: list[type[InterceptorProtocol]]) -> None:
    with _lock:
        for cls in interceptor_classes:
            if cls in _installed:
                instance, count = _installed[cls]
                _installed[cls] = (instance, count + 1)
            else:
                instance = cls()
                instance.install()
                _installed[cls] = (instance, 1)


def release_patches(interceptor_classes: list[type[InterceptorProtocol]]) -> None:
    with _lock:
        for cls in interceptor_classes:
            if cls not in _installed:
                continue
            instance, count = _installed[cls]
            if count <= 1:
                instance.uninstall()
                del _installed[cls]
            else:
                _installed[cls] = (instance, count - 1)
